from typing import Callable, Iterable, Optional, Tuple, Union

import operator as op

from itertools import product
from itertools import starmap

import numpy as np

from nonion import padr
from nonion import replicate
from nonion import slide

Coordinate = Tuple[int, ...]
RawSlice = Union[
  Tuple[Optional[int]],
  Tuple[Optional[int], Optional[int]],
  Tuple[Optional[int], Optional[int], Optional[int]]
]
Shape = Tuple[int, ...]

def slices(xs: Iterable[RawSlice]) -> Tuple[slice, ...]:
  return tuple(starmap(slice, xs))

def batches(
  n: int = 32,
  step: Optional[int] = None,
  exact: bool = True
  ) -> Callable[[Iterable[Tuple[np.ndarray, ...]]], Iterable[Tuple[np.ndarray, ...]]]:

  return lambda xs: map(batch, slide(n, step or n, exact)(xs))

def batch(ys: Tuple[Tuple[np.ndarray, ...], ...]) -> Tuple[np.ndarray, ...]:
  return tuple(map(np.stack, zip(*ys)))

def cycle(tensor: Union[np.ndarray, float], shape: Shape) -> np.ndarray:
  tensor = tensor if isinstance(tensor, np.ndarray) else np.array(tensor)

  s = tuple(padr(len(shape), 1)(tensor.shape))
  t = tensor.reshape(s)

  for axis, n in enumerate(shape):
    t = np.concatenate(tuple(replicate(n)(t)), axis=axis)

  return t

def gridrange(
  start: Coordinate,
  end: Optional[Coordinate] = None,
  step: Optional[Coordinate] = None
  ) -> Iterable[Coordinate]:

  if end is None:
    end = start
    start = tuple(replicate(len(end))(0))

  step = step or tuple(replicate(len(start))(1))

  ranges = map(range, start, end, step)
  return product(*ranges)

def iteraxis(tensor: np.ndarray, axis: int = 0) -> Iterable[np.ndarray]:
  t = np.moveaxis(tensor, axis, -1)
  return map(lambda x: t[x], np.ndindex(t.shape[:-1]))

def move(
  source: np.ndarray,
  destination: np.ndarray,
  coordinate: Optional[Coordinate] = None,
  inplace: bool = False
  ) -> np.ndarray:

  destination = destination if inplace else destination.copy()
  coordinate = coordinate or ()

  start: Coordinate = tuple(padr(source.ndim, 0)(coordinate))
  end = map(op.add, start, source.shape)

  raw_subslices = zip(start, end)
  subslices = slices(raw_subslices)

  destination[subslices] = source

  return destination



from typing import Callable, Dict, Tuple, Union

import warnings

from dataclasses import dataclass
from functools import partial

import numpy as np
import pandas as pd

from nonion import Maybe
from nonion import Pipeline
from nonion import as_catch
from nonion import in_
from nonion import star
from nonion import wraptry

from sklearn.preprocessing import OneHotEncoder

Dummifier = Callable[[pd.DataFrame], pd.DataFrame]

ENCODER_PARAMETERS = {
  "sparse": False,
  "handle_unknown": "ignore"
}

def dummifier(
  xys: Dict[str, Tuple[object, ...]],
  name: Callable[[str, object], str] = lambda n, x: n + "_" + str(x)
  ) -> Dummifier:

  to_encoder = (
    Pipeline(xys.items())
    / star(lambda x, y: (x, encoder(y, partial(name, x))))
    >> as_catch(lambda _: pd.DataFrame)
  )

  def g(xs: pd.DataFrame) -> pd.DataFrame:
    frame = (
      Pipeline(xs.iteritems())
      / star(lambda n, x: to_encoder(n)(x))
      >> partial(pd.concat, axis=1)
    )
    frame.index = xs.index
    return frame

  return g

@dataclass
class RawEncoder:
  encode: Callable[[np.ndarray], Maybe[np.ndarray]]
  invert: Callable[[np.ndarray], Maybe[np.ndarray]]
  categories: np.ndarray
  columns: Tuple[str, ...]
  dtype: type

class BaseEncoder:
  def __init__(self, raw: RawEncoder):
    self._raw = raw

    self.categories = raw.categories
    self.columns = raw.columns
    self.dtype = raw.dtype

class InverseEncoder:
  pass

class Encoder(BaseEncoder):
  def __init__(self, raw: RawEncoder):
    super().__init__(raw)

  def encode(self, xs: Tuple[object, ...]) -> pd.DataFrame:
    ys = np.expand_dims(xs, axis=-1)

    with warnings.catch_warnings():
      warnings.simplefilter("ignore")
      frame, *_ = self._raw.encode(ys) or (np.zeros((len(xs), len(self.columns))),)

    return pd.DataFrame(frame, columns=self.columns)

  def invert(self) -> InverseEncoder:
    return InverseEncoder(self._raw)

  def __call__(self, xs: Tuple[object, ...]) -> pd.DataFrame:
    return self.encode(xs)

  def __invert__(self) -> InverseEncoder:
    return self.invert()

class InverseEncoder(BaseEncoder):
  def __init__(self, raw: RawEncoder):
    super().__init__(raw)

  def encode(self, xs: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
    if isinstance(xs, pd.DataFrame):
      cs = Pipeline(self.columns) % in_(tuple(xs.columns)) >> list
      xs = xs.loc[:, cs].to_numpy()

    k = xs.shape[1] if xs.ndim == 2 else 0

    if k == len(self.columns):
      frame, *_ = self._raw.invert(xs) or (np.empty((0,), dtype=self.dtype),)
      return frame.reshape(-1)
    else:
      return np.empty((0,), dtype=self.dtype)

  def invert(self) -> Encoder:
    return Encoder(self._raw)

  def __call__(self, xs: Tuple[object, ...]) -> pd.DataFrame:
    return self.encode(xs)

  def __invert__(self) -> Encoder:
    return self.invert()

def encoder(xs: Tuple[object, ...], name: Callable[[object], str] = str) -> Encoder:
  e = OneHotEncoder(**ENCODER_PARAMETERS)
  e.fit(np.expand_dims(xs, axis=-1))

  categories, *_ = e.categories_
  raw = RawEncoder(
    encode=wraptry(e.transform),
    invert=wraptry(e.inverse_transform),
    categories=categories,
    columns=tuple(map(name, categories)),
    dtype=e.dtype
  )

  return Encoder(raw)
